/* $Id: kasm_mnemonics.cpp,v 1.55 2001/03/03 11:30:47 ekalin Exp $ */

/*
 *   kasm: Assembler for the Saturn processor, used in HP48 and HP49
 *   graphical calculators.
 *   Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <iostream>
#include <iomanip>
#include <string>
#include <deque>
#include <map>
#include <stdlib.h>


#include "entrytable.hpp"
#include "nib_io.hpp"
#include "kasm.hpp"


/* Global variables */
unsigned long pc = 0;
unsigned long last_test_opcode;
unsigned long last_test_rev;
int last_test_len = 0;
unsigned long last_closed_block;
unsigned long blk_count;
deque<Block_Info> blk_info;
deque<unsigned long> blk_curr;


static Equivtable fields_a[] = {
  { "P",  0 },
  { "WP", 1 },
  { "XS", 2 },
  { "X",  3 },
  { "S",  4 },
  { "M",  5 },
  { "B",  6 },
  { "W",  7 },
  { "A",  0xF },
};


void
mnemaddsub(char *mnem, char *field)
{
  static Equivtable mnemonics[] = {
    { "A=A+B", 0xA00 },
    { "B=B+C", 0xA01 },
    { "C=C+A", 0xA02 },
    { "D=D+C", 0xA03 },
    { "A=A+A", 0xA04 },
    { "B=B+B", 0xA05 },
    { "C=C+C", 0xA06 },
    { "D=D+D", 0xA07 },
    { "B=B+A", 0xA08 },
    { "C=C+B", 0xA09 },
    { "A=A+C", 0xA0A },
    { "C=C+D", 0xA0B },
    { "A=A-B", 0xB00 },
    { "B=B-C", 0xB01 },
    { "C=C-A", 0xB02 },
    { "D=D-C", 0xB03 },
    { "B=B-A", 0xB08 },
    { "C=C-B", 0xB09 },
    { "A=A-C", 0xB0A },
    { "C=C-D", 0xB0B },
    { "A=B-A", 0xB0C },
    { "B=C-B", 0xB0D },
    { "C=A-C", 0xB0E },
    { "D=C-D", 0xB0F },
  };
  unsigned long opcode;
  unsigned int fieldop;
  unsigned int len = 3;

  opcode = lookup(mnemonics, sizeof(mnemonics) / sizeof(Equivtable), mnem);
  if (opcode == 0) {
    yyerror("Invalid register combination");
    return;
  }

  if (strcmp(field, "A") != 0) {
    fieldop = lookup(fields_a, sizeof(fields_a) / sizeof(Equivtable), field);
    fieldop <<= 4;
    opcode |= fieldop;
  } else {
    if ((opcode & 0xF00) == 0xA00)
      opcode |= 0xC0;
    else
      opcode |= 0xE0;
    opcode &= 0xFF;
    len = 2;
  }

  output(opcode, len);
}


void
mnemaddsubcon(char *mnem, char *field, long n)
{
  if (mnem[0] != mnem[2]) {
    yyerror("Invalid register combination");
    return;
  }

  if (n == 0)
    return;
  if (n < 0) {
    n = -n;
    if (mnem[3] == '+')
      mnem[3] = '-';
    else
      mnem[3] = '+';
  }

  unsigned long opcode;
  unsigned int len;

  if (n == 1) {
    if (strcmp(field, "A") == 0) {
      opcode = (mnem[3] == '+') ? 0xE4 : 0xCC;
      len = 2;
    } else {
      opcode = (mnem[3] == '+') ? 0xB04 : 0xA0C;
      unsigned long fieldop = lookup(fields_a,
                                     sizeof(fields_a) / sizeof(Equivtable),
                                     field);
      fieldop <<= 4;
      opcode |= fieldop;
      len = 3;
    }
    opcode += (mnem[0] - 'A');
  } else {
    opcode = (mnem[3] == '+') ? 0x818000 : 0x818080;
    len = 6;
    unsigned long fieldop = lookup(fields_a,
                                   sizeof(fields_a) / sizeof(Equivtable),
                                   field);

    if (fieldop <= 2 || fieldop == 4)   // P WP XS and S fields
      yyerror("warning: Bugged instruction in Saturn processor.");
    
    fieldop <<= 8;
    opcode |= fieldop;
    opcode += (mnem[0] - 'A') * 0x10;

    while (n > 16) {
      output(opcode | 0xF, 6);
      n -= 16;
    }

    opcode |= (n - 1);
  }

  output(opcode, len);
}


void
mnemzeroetal(char *mnem, char *field)
{
  static Equivtable mnemonics[] = {
    { "A=0",    0xA00 },
    { "B=0",    0xA01 },
    { "C=0",    0xA02 },
    { "D=0",    0xA03 },
    { "A=B",    0xA04 },
    { "B=C",    0xA05 },
    { "C=A",    0xA06 },
    { "D=C",    0xA07 },
    { "B=A",    0xA08 },
    { "C=B",    0xA09 },
    { "A=C",    0xA0A },
    { "C=D",    0xA0B },
    { "ABEX",   0xA0C },
    { "BCEX",   0xA0D },
    { "ACEX",   0xA0E },
    { "CDEX",   0xA0F },
    { "ASL",    0xB00 },
    { "BSL",    0xB01 },
    { "CSL",    0xB02 },
    { "DSL",    0xB03 },
    { "ASR",    0xB04 },
    { "BSR",    0xB05 },
    { "CSR",    0xB06 },
    { "DSR",    0xB07 },
    { "A=-A",   0xB08 },
    { "B=-B",   0xB09 },
    { "C=-C",   0xB0A },
    { "D=-D",   0xB0B },
    { "A=-A-1", 0xB0C },
    { "B=-B-1", 0xB0D },
    { "C=-C-1", 0xB0E },
    { "D=-D-1", 0xB0F },
  };
  unsigned long opcode;
  unsigned int fieldop;
  unsigned int len = 3;

  opcode = lookup(mnemonics, sizeof(mnemonics) / sizeof(Equivtable), mnem);
  if (opcode == 0) {
    yyerror("Invalid register combination");
    return;
  }

  if (strcmp(field, "A") != 0) {
    fieldop = lookup(fields_a, sizeof(fields_a) / sizeof(Equivtable), field);
    fieldop += 8;
    fieldop <<= 4;
    opcode |= fieldop;
  } else {
    if ((opcode & 0xF00) == 0xA00)
      opcode |= 0xD0;
    else
      opcode |= 0xF0;
    opcode &= 0xFF;
    len = 2;
  }

  output(opcode, len);
}


void
mnemlogical(char *mnem, char *field)
{
  static Equivtable mnemonics[] = {
    { "A=A&B", 0x0E00 },
    { "B=B&C", 0x0E01 },
    { "C=C&A", 0x0E02 },
    { "D=D&C", 0x0E03 },
    { "B=B&A", 0x0E04 },
    { "C=C&B", 0x0E05 },
    { "A=A&C", 0x0E06 },
    { "C=C&D", 0x0E07 },
    { "A=A!B", 0x0E08 },
    { "B=B!C", 0x0E09 },
    { "C=C!A", 0x0E0A },
    { "D=D!C", 0x0E0B },
    { "B=B!A", 0x0E0C },
    { "C=C!B", 0x0E0D },
    { "A=A!C", 0x0E0E },
    { "C=C!D", 0x0E0F },
  };
  unsigned long opcode;
  unsigned int fieldop;

  opcode = lookup(mnemonics, sizeof(mnemonics) / sizeof(Equivtable), mnem);
  if (opcode == 0) {
    yyerror("Invalid register combination");
    return;
  }

  fieldop = lookup(fields_a, sizeof(fields_a) / sizeof(Equivtable), field);
  fieldop <<= 4;
  opcode |= fieldop;

  output(opcode, 4);
}


void
mnemsrb(long reg, char *field)
{
  if (strcmp(field, "W") == 0) {
    unsigned long opcode = 0x81C + reg;
    output(opcode, 3);
  } else {
    unsigned long opcode = 0x81900 + reg;
    unsigned long fieldop =
      lookup(fields_a, sizeof(fields_a) / sizeof(Equivtable), field);
    fieldop <<= 4;
    opcode |= fieldop;
    output(opcode, 5);
  }
}


void
mnemacld(char *mnem, long n)
{
  if (mnem[1] == 'A')
    output(0x8082, 4);
  else
    output(3, 1);

  char size[3];
  size[0] = mnem[3];
  if (mnem[4] != ')') {
    size[1] = mnem[4];
    size[2] = '\0';
  } else
    size[1] = '\0';
  int len = atoi(size);
  output(len - 1, 1);

  if (len > 8)
    yyerror("warning: Loading more than 8 nibbles");

  unsigned long mask = 0xF;
  for (int i = 1; i < len; ++i) {
    mask <<= 4;
    mask |= 0xF;
  }

  outputreverse(n & mask, len);
}


static Equivtable memory_mnemonics[] = {
  { "DAT0=A", 0x1500 },
  { "DAT1=A", 0x1510 },
  { "A=DAT0", 0x1520 },
  { "A=DAT1", 0x1530 },
  { "DAT0=C", 0x1540 },
  { "DAT1=C", 0x1550 },
  { "C=DAT0", 0x1560 },
  { "C=DAT1", 0x1570 },
};

void
mnemmemoryf(char *mnem, char *field)
{
  unsigned long opcode;
  unsigned int fieldop;
  unsigned int len = 4;

  opcode = lookup(memory_mnemonics,
                  sizeof(memory_mnemonics) / sizeof(Equivtable), mnem);
  if (opcode == 0) {
    yyerror("Invalid register combination");
    return;
  }

  if (strcmp(field, "A") == 0) {
    opcode >>= 4;
    opcode -= 0x10;
    len = 3;
  } else if (strcmp(field, "B") == 0) {
    opcode >>= 4;
    opcode -= 0x8;
    len = 3;
  } else {
    fieldop = lookup(fields_a, sizeof(fields_a) / sizeof(Equivtable), field);
    opcode |= fieldop;
  }

  output(opcode, len);
}


void
mnemmemoryn(char *mnem, long size)
{
  unsigned long opcode;

  if (size > 16 || size < 1) {
    yyerror("Value between 1-16 expected");
    return;
  }

  opcode = lookup(memory_mnemonics,
                  sizeof(memory_mnemonics) / sizeof(Equivtable), mnem);
  if (opcode == 0) {
    yyerror("Invalid register combination");
    return;
  }

  --size;

  opcode |= size;
  opcode += 0x80;
  output(opcode, 4);
}


void
mnembit(long opcode, long n)
{
  if (n < 0 || n > 15) {
    yyerror("Value between 0-15 expected");
    return;
  }

  output(opcode | n, 5);
}


void
mnemdatregadd(char *mnem, long n)
{
  static Equivtable mnemonics[] = {
    { "D0=D0+", 0x160 },
    { "D1=D1+", 0x170 },
    { "D0=D0-", 0x180 },
    { "D1=D1-", 0x1C0 },
  };

  if (n < 0) {                  /* If necessary, change sign. */
    n = -n;
    if (mnem[5] == '-')
      mnem[5] = '+';
    else
      mnem[5] = '-';
  }

  unsigned long opcode = lookup(mnemonics,
                                sizeof(mnemonics) / sizeof(Equivtable), mnem);
  if (opcode == 0) {
    yyerror("Invalid register combination");
    return;
  }

  while (n >= 16) {
    output(opcode | 0xF, 3);
    n -= 16;
  }

  if (n == 0)
    return;

  --n;
  output(opcode | n, 3);
}


void
mnemdatregld(char *mnem, unsigned long n)
{
  unsigned long opcode = 0x19;
  if (mnem[1] == '1')
    opcode += 4;

  n &= 0xFFFFF;         /* Keep only last 5 nibbles. */

  int size = 5;
  if (strlen(mnem) > 3)
    size = mnem[4] - '0';
  else {
    if ((n & 0xFF) == n)
      size = 2;
    else if ((n & 0xFFFF) == n)
      size = 4;
  }

  if (size == 4)
    opcode += 1;
  else if (size == 5)
    opcode += 2;

  output(opcode, 2);
  outputreverse(n, size);
}


void
mnemscratchreg(long op, char *field)
{
  if (strcmp(field, "W") == 0) {
    unsigned long opcode = 0x100 + op;
    output(opcode, 3);
  } else {
    unsigned long opcode = 0x81A000 + op;
    unsigned long fieldop =
      lookup(fields_a, sizeof(fields_a) / sizeof(Equivtable), field);
    fieldop <<= 8;
    opcode |= fieldop;
    output(opcode, 6);
  }
}


void
mnempld(long n)
{
  if (n < 0 || n > 15) {
    yyerror("Value between 0-15 expected");
    return;
  }

  output(0x20 | n, 2);
}


void
mnemcpasn(long opcode, long n)
{
  if (n < 0 || n > 15) {
    yyerror("Value between 0-15 expected");
    return;
  }

  output(opcode | n, 4);
}


void
mnemsetst(long opcode, long n)
{
  if (n < 0 || n > 15) {
    yyerror("Value between 0-15 expected");
    return;
  }

  output(opcode | n, 3);
}


void
mnemgoto(char *mnem, long n)
{
  int len = last_test_len;
  last_test_len = 0;
  if ((last_test_rev == 0) && (len != 0)) {
    yyerror("This test cannot be reversed");
    return;
  }
  
  if (strcmp(mnem, "GOTO") == 0) {
    if (len != 0) {
      output(last_test_rev, len);
      outputreverse(6, 2);      /* Amount to jump */
    }
    
    n -= (pc + 1);
    if (n < -(0x800) || n > 0x7FF)
      yyerror("Jump too long");
    else {
      output(6, 1);
      outputreverse(n & 0xFFF, 3);
    }
  } else if (strcmp(mnem, "GOSUB") == 0) {
    if (len != 0) {
      output(last_test_rev, len);
      outputreverse(6, 2);      /* Amount to jump */
    }
    
    n -= (pc + 4);
    if (n < -(0x800) || n > 0x7FF)
      yyerror("Jump too long");
    else {
      output(7, 1);
      outputreverse(n & 0xFFF, 3);
    }
  } else if (strcmp(mnem, "GOC") == 0 || strcmp(mnem, "GONC") == 0) {
    if (len != 0) {
      yyerror("Forbidden");
      return;
    }
    
    n -= (pc + 1);
    if (n < -(0x80) || n > 0x7F)
      yyerror("Jump too long");
    else {
      output((strcmp(mnem, "GOC") == 0 ? 4 : 5), 1);
      outputreverse(n & 0xFF, 2);
    }
  } else if (strcmp(mnem, "GOLONG") == 0 || strcmp(mnem, "GOTOL") == 0) {
    if (len != 0) {
      output(last_test_rev, len);
      outputreverse(8, 2);      /* Amount to jump */
    }
    
    n -= (pc + 2);
    if (n < -(0x8000) || n > 0x7FFF)
      yyerror("Jump too long.");
    else {
      output(0x8C, 2);
      outputreverse(n & 0xFFFF, 4);
    }
  } else if (strcmp(mnem, "GOSUBL") == 0) {
    if (len != 0) {
      output(last_test_rev, len);
      outputreverse(8, 2);      /* Amount to jump */
    }
    
    n -= (pc + 6);
    if (n < -(0x8000) || n > 0x7FFF)
      yyerror("Jump too long.");
    else {
      output(0x8E, 2);
      outputreverse(n & 0xFFFF, 4);
    }
  } else if (strcmp(mnem, "GOVLNG") == 0) {
    if (len != 0) {
      output(last_test_rev, len);
      outputreverse(9, 2);      /* Amount to jump */
    }
    
    output(0x8D, 2);
    outputreverse(n & 0xFFFFF, 5);
  } else {                      // GOSBVL
    if (len != 0) {
      output(last_test_rev, len);
      outputreverse(9, 2);      /* Amount to jump */
    }
    
    output(0x8F, 2);
    outputreverse(n & 0xFFFFF, 5);
  }
}


void
mnemtest1(char *mnem, char *field)
{
  static Equivtable mnemonics[] = {
    { "?A=B", 0x900 },
    { "?B=C", 0x901 },
    { "?A=C", 0x902 },
    { "?C=D", 0x903 },
    { "?A#B", 0x904 },
    { "?B#C", 0x905 },
    { "?A#C", 0x906 },
    { "?C#D", 0x907 },
    { "?A=0", 0x908 },
    { "?B=0", 0x909 },
    { "?C=0", 0x90A },
    { "?D=0", 0x90B },
    { "?A#0", 0x90C },
    { "?B#0", 0x90D },
    { "?C#0", 0x90E },
    { "?D#0", 0x90F },
  };
  unsigned long opcode;
  unsigned int fieldop;

  opcode = lookup(mnemonics, sizeof(mnemonics) / sizeof(Equivtable), mnem);
  if (opcode == 0) {
    yyerror("Invalid register combination");
    return;
  }

  if (strcmp(field, "A") != 0) {
    fieldop = lookup(fields_a, sizeof(fields_a) / sizeof(Equivtable), field);
    fieldop <<= 4;
    opcode |= fieldop;
  } else {
    opcode -= 0x100;
    opcode |= 0xA0;
  }

  last_test_opcode = last_test_rev = opcode;
  if ((last_test_rev & 0x7) >= 0x4)
    last_test_rev -= 0x4;
  else
    last_test_rev += 0x4;
  last_test_len = 3;
}


void
mnemtest2(char *mnem, char *field)
{
  static Equivtable mnemonics[] = {
    { "?A>B",  0x900 },
    { "?B<=A", 0x900 },
    { "?B>C",  0x901 },
    { "?C<=B", 0x901 },
    { "?C>A",  0x902 },
    { "?A<=C", 0x902 },
    { "?D>C",  0x903 },
    { "?C<=D", 0x903 },
    { "?A<B",  0x904 },
    { "?B>=A", 0x904 },
    { "?B<C",  0x905 },
    { "?C>=B", 0x905 },
    { "?C<A",  0x906 },
    { "?A>=C", 0x906 },
    { "?D<C",  0x907 },
    { "?C>=D", 0x907 },
    { "?A>=B", 0x908 },
    { "?B<A",  0x908 },
    { "?B>=C", 0x909 },
    { "?C<B",  0x909 },
    { "?C>=A", 0x90A },
    { "?A<C",  0x90A },
    { "?D>=C", 0x90B },
    { "?C<D",  0x90B },
    { "?A<=B", 0x90C },
    { "?B>A",  0x90C },
    { "?B<=C", 0x90D },
    { "?C>B",  0x90D },
    { "?C<=A", 0x90E },
    { "?A>C",  0x90E },
    { "?D<=C", 0x90F },
    { "?C>D",  0x90F },
  };
  unsigned long opcode;
  unsigned int fieldop;

  opcode = lookup(mnemonics, sizeof(mnemonics) / sizeof(Equivtable), mnem);
  if (opcode == 0) {
    yyerror("Invalid register combination");
    return;
  }

  if (strcmp(field, "A") != 0) {
    fieldop = lookup(fields_a, sizeof(fields_a) / sizeof(Equivtable), field);
    fieldop += 0x8;
    fieldop <<= 4;
    opcode |= fieldop;
  } else {
    opcode -= 0x100;
    opcode |= 0xB0;
  }

  last_test_opcode = last_test_rev = opcode;
  if ((last_test_rev & 0xF) <= 3)
    last_test_rev += 0xC;
  else if ((last_test_rev & 0xF) <= 7)
    last_test_rev += 0x4;
  else if ((last_test_rev & 0xF) <= 0xB)
    last_test_rev -= 0x4;
  else
    last_test_rev -= 0xC;
  last_test_len = 3;
}


void
mnemtest3(long opcode, long n)
{
  if (n < 0 || n > 15) {
    yyerror("Value between 0-15 expected");
    return;
  }

  last_test_opcode = opcode | n;
  if (last_test_rev != 0)
    last_test_rev |= n;
  last_test_len = 3;
}


void
mnemtest5(long opcode, long n)
{
  if (n < 0 || n > 15) {
    yyerror("Value between 0-15 expected");
    return;
  }

  last_test_opcode = opcode | n;
  last_test_rev |= n;
  last_test_len = 5;
}


void
goyes(long n)
{
  if (last_test_len == 0) {
    yyerror("Forbidden");
    return;
  }

  int len = last_test_len;
  last_test_len = 0;
  
  n -= (pc + len);
  if (n < -(0x80) || n > 0x7F)
    yyerror("Jump too long");
  else {
    output(last_test_opcode, len);
    outputreverse(n & 0xFF, 2);
  }
}


void
mnemnofield2(char *mnem)
{
  static Equivtable mnemonics[] = {
    { "RTNSXM", 0x00 },
    { "RTN",    0x01 },
    { "RTNSC",  0x02 },
    { "RTNCC",  0x03 },
    { "SETHEX", 0x04 },
    { "SETDEC", 0x05 },
    { "RSTK=C", 0x06 },
    { "C=RSTK", 0x07 },
    { "CLRST",  0x08 },
    { "C=ST",   0x09 },
    { "ST=C",   0x0A },
    { "CSTEX",  0x0B },
    { "P=P+1",  0x0C },
    { "P=P-1",  0x0D },
    { "P+1",    0x0C },
    { "P-1",    0x0D },
    { "RTI",    0x0F },
  };
  unsigned long opcode;

  opcode = lookup(mnemonics, sizeof(mnemonics) / sizeof(Equivtable), mnem);
  output(opcode, 2);
}


void
mnemnofield3(char *mnem)
{
  static Equivtable mnemonics[] = {
    { "RTNC",   0x400 },
    { "NOP3",   0x420 },
    { "RTNNC",  0x500 },
    { "R0=A",   0x100 },
    { "R1=A",   0x101 },
    { "R2=A",   0x102 },
    { "R3=A",   0x103 },
    { "R4=A",   0x104 },
    { "R0=C",   0x108 },
    { "R1=C",   0x109 },
    { "R2=C",   0x10A },
    { "R3=C",   0x10B },
    { "R4=C",   0x10C },
    { "A=R0",   0x110 },
    { "A=R1",   0x111 },
    { "A=R2",   0x112 },
    { "A=R3",   0x113 },
    { "A=R4",   0x114 },
    { "C=R0",   0x118 },
    { "C=R1",   0x119 },
    { "C=R2",   0x11A },
    { "C=R3",   0x11B },
    { "C=R4",   0x11C },
    { "AR0EX",  0x120 },
    { "AR1EX",  0x121 },
    { "AR2EX",  0x122 },
    { "AR3EX",  0x123 },
    { "AR4EX",  0x124 },
    { "CR0EX",  0x128 },
    { "CR1EX",  0x129 },
    { "CR2EX",  0x12A },
    { "CR3EX",  0x12B },
    { "CR4EX",  0x12C },
    { "D0=A",   0x130 },
    { "D1=A",   0x131 },
    { "AD0EX",  0x132 },
    { "AD1EX",  0x133 },
    { "D0=C",   0x134 },
    { "D1=C",   0x135 },
    { "CD0EX",  0x136 },
    { "CD1EX",  0x137 },
    { "D0=AS",  0x138 },
    { "D1=AS",  0x139 },
    { "AD0XS",  0x13A },
    { "AD1XS",  0x13B },
    { "D0=CS",  0x13C },
    { "D1=CS",  0x13D },
    { "CD0XS",  0x13E },
    { "CD1XS",  0x13F },
    { "OUT=CS", 0x800 },
    { "OUT=C",  0x801 },
    { "A=IN",   0x802 },
    { "C=IN",   0x803 },
    { "UNCNFG", 0x804 },
    { "CONFIG", 0x805 },
    { "C=ID",   0x806 },
    { "SHUTDN", 0x807 },
    { "C+P+1",  0x809 },
    { "RESET",  0x80A },
    { "BUSCC",  0x80B },
    { "SREQ?",  0x80E },
    { "ASLC",   0x810 },
    { "BSLC",   0x811 },
    { "CSLC",   0x812 },
    { "DSLC",   0x813 },
    { "ASRC",   0x814 },
    { "BSRC",   0x815 },
    { "CSRC",   0x816 },
    { "DSRC",   0x817 },
    { "XM=0",   0x821 },
    { "SB=0",   0x822 },
    { "SR=0",   0x824 },
    { "MP=0",   0x828 },
    { "CLRHST", 0x82F },
  };
  unsigned long opcode;

  opcode = lookup(mnemonics, sizeof(mnemonics) / sizeof(Equivtable), mnem);

  if (opcode == 0x802 || opcode == 0x803)    // A=IN and C=IN
    yyerror("warning: Bugged instruction in Saturn processor.");
  
  output(opcode, 3);
}


void
mnemnofield4(char *mnem)
{
  static Equivtable mnemonics[] = {
    { "NOP4",   0x6300 },
    { "INTON",  0x8080 },
    { "BUSCB",  0x8083 },
    { "PC=(A)", 0x808C },
    { "BUSCD",  0x808D },
    { "PC=(C)", 0x808E },
    { "INTOFF", 0x808F },
    { "PC=A",   0x81B2 },
    { "PC=C",   0x81B3 },
    { "A=PC",   0x81B4 },
    { "C=PC",   0x81B5 },
    { "APCEX",  0x81B6 },
    { "CPCEX",  0x81B7 },
  };
  unsigned long opcode;

  opcode = lookup(mnemonics, sizeof(mnemonics) / sizeof(Equivtable), mnem);
  output(opcode, 4);
}


void
mnemskip(bool isgosub)
{
  if (firstpass) {
    blk_info[blk_curr.back()].start += 4; /* Adjustment so that the SKIP is
                                             not a part of the block. */
    output(0, 4);
    return;
  }

  output(isgosub ? 7 : 6, 1);
  long address = blk_info[blk_curr.back()].end;
  address -= pc;
  if (isgosub)
    address -= 3;
  
  if (address < -(0x800) | address > 0x7FF)
    yyerror("Jump too long");
  outputreverse(address & 0xFFF, 3);
}


void
mnemskip_c(bool iscarry)
{
  if (firstpass) { 
    blk_info[blk_curr.back()].start += 3; /* Adjustment so that the SKIP is
                                             not a part of the block. */
    output(0, 3);
    return;
  }

  output(iscarry ? 4 : 5, 1);
  long address = blk_info[blk_curr.back()].end;
  address -= pc;
  
  if (address < -(0x80) | address > 0x7F)
    yyerror("Jump too long");
  outputreverse(address & 0xFF, 2);
}


void
mnemskip_l(bool isgosub)
{
  if (firstpass) {
    blk_info[blk_curr.back()].start += 6; /* Adjustment so that the SKIP is
                                             not a part of the block. */
    output(0, 6);
    return;
  }

  output(isgosub ? 0x0E : 0x8C, 2);
  long address = blk_info[blk_curr.back()].end;
  address -= pc;
  if (isgosub)
    address -= 4;
  
  if (address < -(0x8000) | address > 0x7FFF)
    yyerror("Jump too long");
  outputreverse(address & 0xFFFF, 4);
}


void
mnemskelse()
{
  /* In the first pass, the type is always NONE, the fake outputting
     is handled in the lexer. */
  switch (blk_info[last_closed_block].skelse_type) {
  case SKELSE:
  {
    long blk_num = blk_info[last_closed_block].else_block;
    long offset = blk_info[blk_num].end - (pc + 1);
    if (offset < -(0x800) || offset > 0x7FF)
      yyerror("Jump too long");
    else {
      output(6, 1);
      outputreverse(offset & 0xFFF, 3);
    }
  }
  break;

  case SKEC:
  case SKENC:
  {
    long blk_num = blk_info[last_closed_block].else_block;
    long offset = blk_info[blk_num].end - (pc + 1);
    if (offset < -(0x80) || offset > 0x7F)
      yyerror("Jump too long");
    else {
      output(blk_info[last_closed_block].skelse_type == SKEC ? 4 : 5, 1);
      outputreverse(offset & 0xFF, 2);
    }
    break;
  }
  }
}


void
mnemexitorup(long n, bool isexit)
{
  int len = last_test_len;
  last_test_len = 0;
  
  if (blk_curr.size() == 0) {
    yyerror("Forbidden outside block");
    return;
  }
  if (blk_curr.size() < (unsigned long)n + 1) {
    yyerror("Trying to go out of too many blocks");
    return;
  }

  if (firstpass) {
    output(0, len ? (2 + len) : 4);
    return;
  }
  
  if (len == 0)
    output(6, 1);
  else
    output(last_test_opcode, len);
  
  long address;
  if (isexit) {
    unsigned long num = blk_curr.back() - n;
    if (blk_info[num].skelse_type != NONE)
      num = blk_info[num].else_block;
    
    address = blk_info[num].end;
  } else
    address = blk_info[blk_curr.back() - n].start;
  address -= pc;
  if (len) {
    if (address < -(0x80) || address > 0x7F)
      yyerror("Jump too long");
    outputreverse(address & 0xFF, 2);
  } else {
    if (address < -(0x800) | address > 0x7FF)
      yyerror("Jump too long");
    outputreverse(address & 0xFFF, 3);  
  }
}



void
mnemexitorup_c(long n, bool isexit, bool iscarry)
{
  if (blk_curr.size() == 0) {
    yyerror("Forbidden outside block");
    return;
  }
  if (blk_curr.size() < (unsigned long)n + 1) {
    yyerror("Trying to go out of too many blocks");
    return;
  }

  if (firstpass) {
    output(0, 3);
    return;
  }

  output(iscarry ? 4 : 5, 1);
  long address;
  if (isexit)
    address = blk_info[blk_curr.back() - n].end;
  else
    address = blk_info[blk_curr.back() - n].end;
  address -= pc;
  if (address < -(0x80) || address > 0x7F)
    yyerror("Jump too long");
  outputreverse(address & 0xFF, 2);  
}


void
mnemstrobj(long prolog)
{
  if (firstpass) {
    blk_info[blk_curr.back()].start += 10; /* Adjustment so that the
                                              prolog/len are not part of
                                              the block. */
    output(0, 10);
    return;
  }

  outputreverse(prolog, 5);
  long address = blk_info[blk_curr.back()].end;
  address -= pc;
  
  outputreverse(address, 5);  
}


void
doptrbint(char *str, bool is_ptr)
{
  char *rest;
  unsigned long value = strtoul(str, &rest, 16);

  if (*rest)
    yyerror("Hexa expected");

  if (!is_ptr) {
    unsigned long addr = get_address_of("DOBINT");
    if (addr == 0) {
      yyerror("Cannot find bint prolog (DOBINT)", true);
      return;
    }
    outputreverse(addr, 5);
  }

  outputreverse(value, 5);
}


void
doId(char *name, bool is_id)
{
  unsigned long addr;
  if (is_id) {
    addr = get_address_of("DOIDNT");
    if (addr == 0) {
      yyerror("Cannot find id prolog (DOIDNT)", true);
      return;
    }
  } else {
    addr = get_address_of("DOLAM");
    if (addr == 0) {
      yyerror("Cannot find lam prolog (DOLAM)", true);
      return;
    }
  }
  
  outputreverse(addr, 5);
  if (!firstpass)
    fOut.putIdNameForm(name);
  pc += strlen(name)*2 + 2;
}


void
dostring(char *str)
{
  unsigned long addr = get_address_of("DOCSTR");
  if (addr == 0) {
    yyerror("Cannot find string prolog (DOCSTR)", true);
    return;
  }

  outputreverse(addr, 5);
  if (!firstpass)
    fOut.putString(str);
  pc += strlen(str)*2 + 5;
}


void
dochar(unsigned char chr)
{
  unsigned long addr = get_address_of("DOCHAR");
  if (addr == 0) {
    yyerror("Cannot find character prolog (DOCHAR)", true);
    return;
  }

  outputreverse(addr, 5);
  if (!firstpass)
    fOut.putChar(chr);
  pc += 2;
}


void
doreal(char *str)
{
  char *rest;
  double n = strtod(str, &rest);
  if (*rest) {
    yyerror("Invalid number");
    return;
  }

  unsigned long addr = get_address_of("DOREAL");
  if (addr == 0) {
    yyerror("Cannot find real number prolog (DOREAL)", true);
    return;
  }

  outputreverse(addr, 5);
  fOut.putReal(n);
  pc += 16;
}


void
doexreal(char *str)
{
  char *rest;
  double n = strtod(str, &rest);
  if (*rest) {
    yyerror("Invalid number");
    return;
  }

  unsigned long addr = get_address_of("DOEREL");
  if (addr == 0) {
    yyerror("Cannot find extended real number prolog (DOEREL)", true);
    return;
  }

  outputreverse(addr, 5);
  fOut.putExReal(n);
  pc += 21;
}


void
docmp(char *re_str, char *im_str)
{
  char *rest;
  double re = strtod(re_str, &rest);
  if (*rest) {
    yyerror("Invalid number");
    return;
  }
  double im = strtod(im_str, &rest);
  if (*rest) {
    yyerror("Invalid number");
    return;
  }

  unsigned long addr = get_address_of("DOCMP");
  if (addr == 0) {
    yyerror("Cannot find complex number prolog (DOCMP)", true);
    return;
  }

  outputreverse(addr, 5);
  fOut.putReal(re);
  fOut.putReal(im);
  pc += 32;
}


void
doexcmp(char *re_str, char *im_str)
{
  char *rest;
  double re = strtod(re_str, &rest);
  if (*rest) {
    yyerror("Invalid number");
    return;
  }
  double im = strtod(im_str, &rest);
  if (*rest) {
    yyerror("Invalid number");
    return;
  }

  unsigned long addr = get_address_of("DOECMP");
  if (addr == 0) {
    yyerror("Cannot find extended complex number prolog (DOECMP)", true);
    return;
  }

  outputreverse(addr, 5);
  fOut.putExReal(re);
  fOut.putExReal(im);
  pc += 42;
}


void
dotag(char *tag)
{
  unsigned long addr = get_address_of("DOTAG");
  if (addr == 0) {
    yyerror("Cannot find tagged object prolog (DOTAG)", true);
    return;
  }

  outputreverse(addr, 5);
  if (!firstpass)
    fOut.putIdNameForm(tag);
  pc += strlen(tag)*2 + 2;
}


void
doromptr(char *str1, char *str2)
{
  char *rest;

  unsigned long value1 = strtoul(str1, &rest, 16);
  if (*rest)
    yyerror("Hexa expected");

  unsigned long value2 = strtoul(str2, &rest, 16);
  if (*rest)
    yyerror("Hexa expected");

  unsigned long addr = get_address_of("DOROMP");
  if (addr == 0) {
    yyerror("Cannot find flashpointer prolog (DOFLASHP)", true);
    return;
  }
  outputreverse(addr, 5);

  outputreverse(value1, 3);
  outputreverse(value2, 3);
}


void
doflashptr(char *str1, char *str2)
{
  char *rest;

  unsigned long value1 = strtoul(str1, &rest, 16);
  if (*rest)
    yyerror("Hexa expected");

  unsigned long value2 = strtoul(str2, &rest, 16);
  if (*rest)
    yyerror("Hexa expected");

  unsigned long addr = get_address_of("DOFLASHP");
  if (addr == 0) {
    yyerror("Cannot find flashpointer prolog (DOFLASHP)", true);
      return;
  }
  outputreverse(addr, 5);

  outputreverse(value1, 3);
  outputreverse(value2, 4);
}


void
doacptr(char *str1, char *str2)
{
  char *rest;

  unsigned long value1 = strtoul(str1, &rest, 16);
  if (*rest)
    yyerror("Hexa expected");

  unsigned long value2 = strtoul(str2, &rest, 16);
  if (*rest)
    yyerror("Hexa expected");

  unsigned long addr = get_address_of("DOACPTR");
  if (addr == 0) {
    yyerror("Cannot find access pointer prolog (DOACPTR)", true);
      return;
  }
  outputreverse(addr, 5);

  outputreverse(value1, 5);
  outputreverse(value2, 5);
}


void
dolenbody(char *prolog, char *len_str, char *body)
{
  unsigned long addr = get_address_of(prolog);
  if (addr == 0) {
    string errortext("Cannot find ");
    errortext += prolog;
    yyerror(errortext.c_str(), true);
      return;
  }

  char *rest;
  unsigned long len = strtoul(len_str, &rest, 16);
  if (*rest) {
    yyerror("Hexa expected");
    return;
  }
  if (len != strlen(body)) {
    yyerror("Invalid size");
    return;
  }

  if (strspn(body, "0123456789ABCDEF") != strlen(body)) {
    yyerror("Hexa expected");
    return;    
  }
  
  outputreverse(addr, 5);
  outputreverse(len + 5, 5);

  for (unsigned int i = 0; i < strlen(body); ++i) {
    if (body[i] >= 'A')
      output(body[i] - 'A' + 10, 1);
    else
      output(body[i] - '0', 1);
  }
}


void
donibb(char *len_str, char *body)
{
  if (len_str != 0) {
    char *rest;
    unsigned long len = strtoul(len_str, &rest, 16);
    if (*rest) {
      yyerror("Hexa expected");
      return;
    }
    if (len != strlen(body)) {
      yyerror("Invalid size");
      return;
    }
  }
  
  if (strspn(body, "0123456789ABCDEF") != strlen(body)) {
    yyerror("Hexa expected");
    return;    
  }
  
  for (unsigned int i = 0; i < strlen(body); ++i) {
    if (body[i] >= 'A')
      output(body[i] - 'A' + 10, 1);
    else
      output(body[i] - '0', 1);
  }
}


void
dozint(char *str)
{
  bool negative = false;
  if (str[0] == '-') {
    negative = true;
    ++str;
  }
  
  if (strspn(str, "0123456789") != strlen(str)) {
    yyerror("Hexa expected");
    return;
  }

  unsigned long addr = get_address_of("DOINT");
  if (addr == 0) {
    yyerror("Cannot find integer prolog (DOINT)", true);
    return;
  }
  outputreverse(addr, 5);

  outputreverse(strlen(str) + 6, 5);
  
  for (int i = strlen(str) - 1; i >= 0; --i)
    output(str[i] - '0', 1);

  if (negative)
    output(9, 1);
  else
    output(0, 1);
}


unsigned long
lookup(Equivtable* table, int n, char *mnem)
{
  int i;
  for (i = 0; i < n; ++i)
    if (strcmp(table[i].mnem, mnem) == 0)
      return table[i].opcode;

  return 0;
}


unsigned long
get_address_of(const string& entry)
{
  Entry_Table_iter i = file_entries.find(entry);
  if (i == file_entries.end()) {
    i = global_entries.find(entry);
    if (i == global_entries.end())
      return 0;
  }
  return i->second->get_address();
}
