/* $Id: main.cpp,v 1.23 2001/03/06 00:50:34 ekalin Exp $ */

/*
 *   kasm: Assembler for the Saturn processor, used in HP48 and HP49
 *   graphical calculators.
 *   Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <iostream>
#include <string>
#include <deque>
#include <map>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>

#include "entrytable.hpp"
#include "nib_io.hpp"
#include "kasm.hpp"


/* Global yacc/lex variables */
extern int yydebug;
extern FILE *yyin;

/* Global variables */
string filename = "stdin";
bool firstpass = true;
Entry_Table file_entries;
Entry_Table global_entries;
nibfstream fOut;
bool ishp49 = IS_HP49_DEFAULT;

/* File global variables */
static int return_code = 0;
static bool no_warnings = false;


int
main(int argc, char *argv[])
{
  string entry_table_file(SHAREDIR DEFAULT_ENTRY_FILE);
  string outfile;
  string header = DEFAULT_HEADER;

  /* For command-line parsing. */
  static struct option long_options[] = {
    { "hp48", no_argument, 0, '8' },
    { "hp49", no_argument, 0, '9' },
    { "header", required_argument, 0, 'h' },
    { "debug", no_argument, 0, 'd' },
    { "output", required_argument, 0, 'o' },
    { "entry-table", required_argument, 0, 't' },
    { "no-warnings", no_argument, 0, 'w' },
    { "version", no_argument, 0, 1 },
    { "help", no_argument, 0, 1 },
    { 0, 0, 0, 0 }
  };
  int long_opt_index = 0;
  
  cout << COPYRIGHT << endl;

  /* Process command line options. */
  opterr = 0;                   /* Display own error message. */
  int opt;
  while ((opt = getopt_long(argc, argv, ":89dh:o:t:w",
                            long_options, &long_opt_index)) != -1)
    switch (opt) {
    case '8':
      entry_table_file = SHAREDIR "/entry48.a";
      header = "HPHP48-X";
      ishp49 = false;
      break;

    case '9':
      entry_table_file = SHAREDIR "/entry49.a";
      header = "HPHP49-C";
      ishp49 = true;
      break;

    case 'h':
      header = optarg;
      break;
      
    case 'd':
      yydebug = 1;
      break;

    case 'o':
      outfile = optarg;
      break;

    case 't':
      entry_table_file = optarg;
      break;

    case 'w':
      no_warnings = true;
      break;

    case 1:  /* A long option with no short equivalent. */
      switch (long_opt_index) {
      case 7:  /* --version */
        /* Version has already been printed. */
        cout << HOMEPAGE;
        exit(0);

      case 8: /* --help */
        cout << HOMEPAGE;
        cout << HELPTEXT;
        exit(0);
      }
      break;
      
    case ':':
      cerr << "kasm: Option -" << optopt << "requires an argument.\n";
      exit(2);

    case '?':
      cerr << "kasm: Unknown option: -" << optopt << ".\n";
      exit(2);
    }

  /* The remaining argument should be the file name. */
  if (argc - optind != 1) {
    cerr << "kasm: Must specify one file.\n";
    exit(2);
  }
  FILE *fp;
  if ((fp = fopen(argv[optind], "r")) == NULL) {
    cerr << "kasm: Cannot open " << argv[optind] << ": "
         << strerror(errno) << endl;
    exit(3);
  }
  yyin = fp;
  filename = argv[optind];

  if (!read_entry_table(entry_table_file, global_entries))
    cerr << "kasm: Cannot open " << entry_table_file << ": "
         << strerror(errno) << endl;

  yyparse();
  if (return_code == 0) {       /* If no errors, make a second pass. */
    firstpass = false;
    pc = 0;
    blk_count = 0;

    if (outfile == "") {
      outfile = argv[optind];
      if ((outfile[outfile.length() - 1] == 's'
           || outfile[outfile.length() - 1] == 'S')
          && outfile[outfile.length() - 2] == '.')
        outfile[outfile.length() - 1] = 'o';
      else
        outfile += ".o";
    }

#ifndef MSDOS
    fOut.open(outfile.c_str(), ios::out | ios::trunc | ios::bin);
#else
    fOut.open(outfile.c_str(), ios::out | ios::trunc | ios::binary);
#endif
    if (!fOut) {
      cerr << "kasm: Cannot open " << outfile << ": "
           << strerror(errno) << endl;
      exit(2);
    }  

    fOut << header;
    
    yyparse();

    fOut.close();
  }

  return return_code;
}



void
output(unsigned long opcode, unsigned int nibbles)
{
  if (last_test_len) {
    yyerror("Forbidden");
    return;
  }

  pc += nibbles;

  last_test_len = 0;
  use_reversed = false;

  if (firstpass)
    return;

  for (int i = nibbles - 1; i >= 0; --i)
    fOut.putNibble((opcode & (0xF << i*4)) >> i*4);
}


void
outputreverse(unsigned long opcode, unsigned int nibbles)
{
  for (unsigned int i = 0; i < nibbles; ++i)
    output((opcode & (0xF << 4*i)) >> 4*i, 1);
}


int
yywrap()
{
  return 1;
}


void
yyerror(const char *s, bool force)
{
  if (strncmp(s, "warning: ", 9) != 0)
    return_code = 1;
  else
    if (no_warnings)
      return;

  if (force || firstpass)
    cerr << filename << ':' << yylineno << ": " << s << endl;
}
