/* $Id: nib_io.cpp,v 1.5 2001/02/08 18:31:17 ekalin Exp $ */

/*
 *   kasm: Assembler for the Saturn processor, used in HP48 and HP49
 *   graphical calculators.
 *   Copyright (C) 2001 Eduardo M Kalinowski (ekalin@iname.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <math.h>

#include "nib_io.hpp"


void
nibfstream::putString(char *str)
{
  /*
   * Writes a string: 5-nibble length field followed by the characters.
   */

  int size = strlen(str);
  put5Nibbles(2*size + 5);

  for (int i = 0; i < size; ++i)
    putChar(str[i]);  
}


void
nibfstream::putIdNameForm(char *str)
{
  /*
   * Writes strings in IDNAMEFORM format: the character count (2 nibbles),
   * followed by the characters.
   */

  unsigned char size = (unsigned char)strlen(str);
  putChar(size);

  for (int i = 0; i < size; ++i)
    putChar(str[i]);
}


void
nibfstream::putReal(double n)
{
  /*
   * Writes a real number:
   * 3-nibble exponent field: for positive exponents, the exponent.
   *    For negative exponents, 1000-exponent.
   * 12-nibble mantissa.
   * 1-nibble sign field: 0 for positive, 9 for negative.
   */

  bool negative = n < 0;
  if (negative)
    n = -n;

  int xp = (int)(n != 0 ? log10(fabs(n)) : 0);
  double mant = (n != 0 ? (n / pow(10, xp)) : 0);
  if (xp < 0)
    xp = 1000 + xp;

  putNibble(xp % 10);
  xp /= 10;
  putNibble(xp % 10);
  xp /= 10;
  putNibble(xp % 10);

  for (int i = 11; i >= 0; --i)
    putNibble((int)fmod(floor(mant * pow(10, i)), 10));

  if (negative)
    putNibble(9);
  else
    putNibble(0);
}


void
nibfstream::putExReal(double n)
{
  /*
   * Writes an extended real number:
   * 5-nibble exponent field: for positive exponents, the exponent.
   *    For negative exponents, 100000-exponent.
   * 15-nibble mantissa.
   * 1-nibble sign field: 0 for positive, 9 for negative.
   */

  bool negative = n < 0;
  if (negative)
    n = -n;

  int xp = (int)(n != 0 ? log10(fabs(n)) : 0);
  double mant = (n != 0 ? (n / pow(10, xp)) : 0);
  if (xp < 0)
    xp = 100000 + xp;

  putNibble(xp % 10);
  xp /= 10;
  putNibble(xp % 10);
  xp /= 10;
  putNibble(xp % 10);
  xp /= 10;
  putNibble(xp % 10);
  xp /= 10;
  putNibble(xp % 10);

  for (int i = 14; i >= 0; --i)
    putNibble((int)fmod(floor(mant * pow(10, i)), 10));

  if (negative)
    putNibble(9);
  else
    putNibble(0);
}


#if 0
unsigned long
nibfstream::get3Nibbles()
{
  /*
   * Gets three nibbles, joining them in reverse order.
   */

  unsigned long a = 0;
  a |= getNibble();
  a |= getNibble() << 4;
  a |= getNibble() << 8;
  if (eof())
    throw "Unexpected end of file";

  return a;
}


unsigned long
nibfstream::get4Nibbles()
{
  /*
   * Gets four nibbles, joining them in reverse order.
   */

  unsigned long a = 0;
  a |= getNibble();
  a |= getNibble() << 4;
  a |= getNibble() << 8;
  a |= getNibble() << 12;
  if (eof())
    throw "Unexpected end of file";

  return a;
}  


unsigned long
nibfstream::get5Nibbles()
{
  /*
   * Reads five nibbles, and returns them joined in inverse order,
   * to form an HP address.
   */

  unsigned long a = 0;
  a |= getNibble();
  a |= getNibble() << 4;
  a |= getNibble() << 8;
  a |= getNibble() << 12;
  a |= getNibble() << 16;
  if (eof())
    throw "Unexpected end of file";

  return a;
}


string
nibfstream::getString(unsigned long &len) 
{
  /*
   * Reads a string. First, the 5-nibble is read (and stored in len),
   * then the characters are read. Throws "Unexpected end of file" if
   * EOF is reached.
   */

  len = get5Nibbles();

  int a = ((int)len - 5) / 2;
  string str = string();

  register int i;
  for (i = 0; i < a; i++)
    str += getTranslationFor(getChar(), true);
    
  if (eof())
    throw "Unexpected end of file";

  return str;
}


string
nibfstream::getIdNameForm(bool translate)
{
  /*
   * Reads strings in IDNAMEFORM format: the character count (2 nibbles),
   * followed by the characters.
   *
   * Throws "Unexpected end of file" if EOF is reached.
   */

  unsigned char n = getChar();

  string str = string();

  unsigned char c;
  for (register int i = 0; i < n; i++) {
    c = getChar();
    if (translate)
      str += getTranslationFor(c, false);
    else
      str += c;
  }

  if (eof())
    throw "Unexpected end of file";

  return str;
}


void
nibfstream::setPos(long pos)
{
  /*
   * Sets the position in the file.
   */

  ibuff = false;
  seekg(pos / 2, ios::beg);
  if (pos & 1)
    (void) getNibble();
}


string
nibfstream::getTranslationFor(unsigned char c, bool expQuote)
{
  /*
   * Returns the string representing the character c. Non-printable
   * characters have their codes output.
   */

  char buff[5] = { '\0' };
  char num[6];

  if (32 <= c && c < 127 && c != 92 && (!expQuote || c != 34))
    buff[0] = c;
  else {
    buff[0] = '\\';
    switch (c) {
    case 9:
      buff[1] = 't';
      break;

    case 10:
      buff[1] = 'n';
      break;

    case 12:
      buff[1] = 'f';
      break;

    case 13:
      buff[1] = 'r';
      break;

    case 34:
      buff[0] = buff[1] = '\"';
      break;

    case 92:
      buff[1] = '\\';
      break;
      
    default:
      sprintf(num, "%02X", (unsigned int)c);
      strcat(buff, num);
      break;
    }
  }

  return string(buff);
}
#endif


void
nibfstream::put5Nibbles(unsigned long a)
{
  /*
   * Puts the HP address, nibble by nibble.
   */

  putNibble((a & 0x0000F));
  putNibble((a & 0x000F0) >> 4);
  putNibble((a & 0x00F00) >> 8);
  putNibble((a & 0x0F000) >> 12);
  putNibble((a & 0xF0000) >> 16);
}


#if 0
void
nibfstream::copyNibblesFrom(nibfstream &fIn, unsigned int len)
{
  /*
   * Copies len nibbles from fIn into the object.
   */

  while (len--)
    putNibble(fIn.getNibble());
}
#endif
